package CPAN::Smoke::Module;

use strict;
use vars qw($VERSION @ISA);

$VERSION = 0.01;

use Archive::Any;
use Carp::Assert;
use CLASS;
use Fatal qw(chdir mkdir);
use File::Spec::Functions;

require Class::Accessor;
@ISA = qw(Class::Accessor);

CLASS->mk_accessors(qw(_fetched_to));


=head1 NAME

CPAN::Smoke::Module - representations of CPAN modules

=head1 SYNOPSIS

  use CPAN::Smoke::Module;

  my $module = CPAN::Smoke::Module->create(\%module_data);
  $module->which_mirror($cpan_mirror);
  $module->fetch($to_some_dir);
  $module->unpack($to_some_other_dir);
  $module->which_perl($perl);
  $module->prefix($dir);
  $module->makefile_pl(@args);
  $module->make;
  $module->make_other($target);
  $module->test;
  $module->install;
  $module->clean;

  my %namespaces = $module->namespaces;
  $module->add_namespace($namespace => $version);

  my %prereqs = $module->prereqs;

  # Simple accessors
  $module->author;
  $module->file_path;
  $module->file;


=head1 DESCRIPTION

A representation of a CPAN module for the purposes of smoking it.

A CPAN module is that which is encased in a single archive (tarball or
zip file) and is typically made up of one or more .pm files with tests
spanning one more or packages.

These objects can do useful things like download, unpack and build
themselves.


=head2 Constructors

=over 4

=item B<create>

  my $module = CPAN::Smoke::Module->create(\%module_data);

Creates a new module object with the given data.

=cut

sub create {
    my($class, $data) = @_;

    my $self = $class->SUPER::new($data);

    $self->which_perl($^X) unless $self->which_perl;

    return $self;
}

=back

=head2 Information about the module.

=item B<author>

  $module->author($author);
  my $author = $module->author;

Author of the module, as CPAN::Smoke::Author object.

=cut

CLASS->mk_accessors('author');

=item B<file>

  $module->file($file);
  my $file = $module->file;

The file this module comes in (like 'Sex-0.69.tar.gz').

=cut

CLASS->mk_accessors('file');

=item B<filepath>

  my $filepath = $module->filepath;

The full CPAN file path to this module (like
'M/MS/MSCHWERN/Sex-0.69.tar.gz')

=cut

sub filepath {
    my($self) = shift;

    return $self->author->dir . '/' . $self->file;
}
   

=item B<namespaces>

  my %namespaces = $module->namespaces;

The namespaces contained in this module and their versions.

    namespace => version

=cut

CLASS->mk_accessors('_namespaces');

sub namespaces {
    my($self) = shift;

    return %{$self->_namespaces};
}   

=item B<add_namespace>

  $module->add_namespace($namespace => $version);

Add a namespace, with it's version, to this module.

=cut

sub add_namespace {
    my($self, $new_space, $version) = @_;
    my $namespaces = $self->_namespaces;

    warn "$new_space already exists" if exists $namespaces->{$new_space};

    $namespaces->{$new_space} = $version;

    return 1;
}


=head2 Unpacking and building

=over 4

=item B<which_mirror>

  $module->which_mirror($cpan_mirror);

What CPAN mirror should we fetch this module from?  This is a full URL
like 'http://www.cpan.org/CPAN'.

You have to set this before you fetch().

=cut

CLASS->mk_accessors('which_mirror');

=item B<fetch>

  my $filepath = $module->fetch($dir);

Goes out and grabs a copy of this module from CPAN (using the entry
from $module->which_mirror) and puts it in the given directory.  If no 
directory is given, it will use the Cwd.

Returns the $filepath to the download.

=cut

use LWP::Simple;

sub fetch {
    my($self, $dir) = @_;

    $dir = curdir() unless defined $dir;

    $self->mistake("Fetch Dir Doesn't Exist", $dir) unless -x $dir;

    my $url = join '/', $self->which_mirror, $self->file_path;

    my $file = catfile($dir, $self->file);
    getstore($url, $file) or $self->mistake("Can't Fetch Module", 
                                            "$url -> $file");

    $self->_fetched_to($dir);

    return $file;
}

=item B<unpack>

  my $unpacked_dir = $module->unpack($dir);

Unpacks the given module into the given directory.  If no directory is
given, it will unpack it in the same directory it was fetched into.

Returns the directory the module unpacked to.

=cut

use Archive::Any;
use Cwd;

sub unpack {
    my($self, $dir) = @_;

    my $new_dir;

    $dir = $self->_fetched_to unless defined $dir;

    my $orig_dir = cwd;
    chdir $dir;

    my $archive = Archive::Any->new($self->file);
    if( $archive->is_naughty ) {
        $self->mistake('Naughty Archive', $self->file);
        return;
    }

    if( $archive->is_impolite ) {
        $self->mistake('Impolite Archive', $self->file);
        
        $new_dir = $self->author->id . $$ . time;
        mkdir $new_dir;
        chdir $new_dir;
        $archive->extract or $self->mistake("Bad Archive", $self->file);
        chdir updir;
    }
    else {
        $new_dir = ($archive->files)[0];
        $archive->extract or $self->mistake("Bad Archive", $self->file);
    }

    chdir $orig_dir;

    $self->_unpacked_to($new_dir);

    return $new_dir;
}

=item B<which_perl>

  my $perl = $module->which_perl;
  $module->which_perl($perl);

Get/set which perl should be used to build & test this module.
If not set, it defaults to the currently running perl.

=cut

CLASS->mk_accessors('which_perl');

=item B<prefix>

  my $prefix = $module->prefix;
  $module->prefix($prefix);

Get/set which directory the module gets installed into (thing PREFIX).

=cut

CLASS->mk_accessors('prefix');

=item B<makefile_pl>

  $module->makefile_pl(@args);

Runs the Makefile.PL against this module using $module->which_perl.

Any arguments are passed to the Makefile.PL on the command line.

=cut

use IPC::Open3;
use IO::Select;

sub makefile_pl {
    my($self, @args) = @_;

    my $orig_dir = cwd;
    chdir($self->_unpacked_to);

    $self->mistake("No Makefile.PL") unless -r 'Makefile.PL';

    my $perl = $self->which_perl;
    my $prefix = $self->prefix;
    my $lib = catdir($prefix, 'lib/');

    my $output = '';
    eval {
        local $SIG{ALRM} = sub { die "ALARM!\n"; };
        alarm 30;

        my($write, $read, $error);
        my $pid = open3($write, $read, $error, $perl, 
                        "Makefile.PL", "LIB=$lib", "PREFIX=$prefix", @args);

        $self->mistake("Couldn't Run Makefile.PL", $!) unless $pid;
        close $write;

        my $select = IO::Select->new($read, $error);
        while( my @fhs = $select->can_read ) {
            foreach my $fh (@fhs) {
                $output = join '', <$fh>;
            }
        }

        alarm 0;
    };
    if( $@ ) {
        if( $@ eq "ALARM!\n" ) {
            $self->mistake("Makefile.PL timed out", $output);
        }
        elsif( ref $@ && $@->isa('CPAN::Mistake::Exception') ) {
            # Propogate CPAN::Mistake exceptions.
            die $@;
        }
        else {
            $self->mistake("UNKNOWN", $output);
        }
    }

    chdir($orig_dir);
}


1;

